package net.sf.javaprinciples.data.transformer;

import java.util.ArrayList;
import java.util.List;

import org.omg.uml.Property;

import au.com.sparxsystems.Association;
import net.sf.jcc.model.parser.UnexpectedException;
import net.sf.jcc.model.parser.uml2.ElementStore;
import net.sf.jcc.model.parser.uml2.ModelElement;

/**
 * Inject the remainder of the dependencies for an AttributeMapper.
 *
 * 1. Construct the set of individual mappings
 *
 * @author Warwick Slade
 */
public class AttributeMapperBuilder<M extends AttributeMapper, T,U> extends ModelElementFactory implements ModelElementMapperBuilder<M>
{
    private ModelElementMapperBuilderBeanFactory modelElementMapperBuilderBeanFactory;

    @Override
    public void build(M mapperModelElement, TransformerFactory.DIRECTION_TYPE direction)
    {
        List<Mapper<T,U>> attributeMapperList = createAttributeMapperList(mapperModelElement.store, mapperModelElement.getModelElement(), direction);
        mapperModelElement.setModelElementMappers(attributeMapperList);
    }

    private List<Mapper<T,U>> createAttributeMapperList(ElementStore store, ModelElement mapperModelElement,
                                                        TransformerFactory.DIRECTION_TYPE direction)
    {
        ArrayList<Mapper<T,U>> attributeMapperList = new ArrayList<Mapper<T,U>>();

        List<Property> mapperModelAttributes = ((org.omg.uml.Class)mapperModelElement.getElement()).getOwnedAttribute();
        for (Property mapperProperty : mapperModelAttributes)
        {
            if (mapperProperty.getAggregation() != null)
            {
                // We are not interested in attributes that actually are an association.
                continue;
            }
            ModelElement mapperPropertyElement = store.reverse(mapperProperty);
            List<Association> mapperPropertyAssociations = mapperPropertyElement.getAssociations();

            List<ModelElement> sourceModelElements = new ArrayList<ModelElement>();
            List<ModelElement> destinationModelElements = new ArrayList<ModelElement>();

            if (mapperPropertyAssociations != null)
            {
                populateSourceAndDestinationModelElements(store, direction, mapperPropertyAssociations, sourceModelElements, destinationModelElements);
            }

            String mapperImplementation = ModelElementMapperHelper.retrieveModelElementTagValue(mapperPropertyElement, IMPLEMENTATION_TAG);
            if (mapperImplementation == null)
            {
                mapperImplementation = "mapper";
            }

            ModelElementMapper<T,U> attributeMapper = getModelElementMapper(mapperImplementation);
            attributeMapper.setStore(store);
            attributeMapper.setModelElement(mapperPropertyElement);

            attributeMapper.setSourceModelElements(sourceModelElements);
            attributeMapper.setDestinationModelElements(destinationModelElements);
            attributeMapperList.add((Mapper)attributeMapper);

            ModelElementMapperBuilder builder = modelElementMapperBuilderBeanFactory.createModelElementMapperBuilder(mapperImplementation + ".builder");
            builder.build(attributeMapper, direction);

//            if (attributeMapper instanceof DefaultAttributeMapper)
//            {
//                if(destinationModelElements.isEmpty())
//                {
//                    throw new UnexpectedException("You must have at least one target model element for %s"
//                            + mapperPropertyElement.getGuid());
//                }
//
//                String leftAttributeMapperDefault = ModelElementMapperHelper.retrieveModelElementTagValue(mapperModelElement, determineDirection(true, direction));
//                ModelElementSourceDestinationMapper sourceAttributeMapper = getModelElementSourceDestinationAttributeMapper(
//                        mapperPropertyElement, determineDirection(true, direction), leftAttributeMapperDefault);
//                ((DefaultAttributeMapper)attributeMapper).setSourceModelElementMapper(sourceAttributeMapper);
//
//                String rightAttributeMapperDefault = ModelElementMapperHelper.retrieveModelElementTagValue(mapperModelElement, determineDirection(false, direction));
//                ModelElementSourceDestinationMapper destinationAttributeMapper = getModelElementSourceDestinationAttributeMapper(
//                        mapperPropertyElement, determineDirection(false, direction), rightAttributeMapperDefault);
//                ((DefaultAttributeMapper)attributeMapper).setDestinationModelElementMapper(destinationAttributeMapper);
//            }


            String sourceModelElementGuid = sourceModelElements.isEmpty() ? null : sourceModelElements.get(0).getGuid();
            String destinationModelElementGuid = destinationModelElements.isEmpty() ? null : destinationModelElements.get(0).getGuid();
            log.info("Creating Mapper: {}, {}, Source Model Element: {}, Target Model Element: {} ",
                    new Object[] {attributeMapper.getClass().getName(), mapperPropertyElement.getGuid() ,
                            sourceModelElementGuid, destinationModelElementGuid});
        }

        if (attributeMapperList.isEmpty())
        {
            throw new UnexpectedException(String.format("Unable to build attributeMapper %s . " +
                    "No associated mappers found.", mapperModelElement.getGuid()));
        }

        return attributeMapperList;
    }

    private void populateSourceAndDestinationModelElements(ElementStore store, TransformerFactory.DIRECTION_TYPE direction, List<Association> mapperPropertyAssociations, List<ModelElement> sourceModelElements, List<ModelElement> destinationModelElements)
    {
        for (Association association : mapperPropertyAssociations)
        {
            String associationStereotype = association.getProperties().getStereotype();
            String targetId = association.getTargetAttributeId() != null ? association.getTargetAttributeId() : association.getTargetClassId();
            ModelElement attributeModelElement = store.get(targetId);

            if (TransformerFactory.DIRECTION_TYPE.leftToRight.equals(direction))
            {
                if (RIGHT_STEREOTYPE_PROPERTY.equals(associationStereotype))
                {
                    destinationModelElements.add(attributeModelElement);
                }
                else if (LEFT_STEREOTYPE_PROPERTY.equals(associationStereotype))
                {
                    sourceModelElements.add(attributeModelElement);
                }
            }
            else
            {
                if (LEFT_STEREOTYPE_PROPERTY.equals(associationStereotype))
                {
                    destinationModelElements.add(attributeModelElement);
                }
                else if (RIGHT_STEREOTYPE_PROPERTY.equals(associationStereotype))
                {
                    sourceModelElements.add(attributeModelElement);
                }
            }
        }
    }

    protected ModelElementMapper<T,U>  createModelElementMapper(String mapperGuid, ElementStore store,
                                                                TransformerFactory.DIRECTION_TYPE direction)
    {
        throw new UnsupportedOperationException();
    }

    public void setModelElementMapperBuilderBeanFactory(ModelElementMapperBuilderBeanFactory modelElementMapperBuilderBeanFactory)
    {
        this.modelElementMapperBuilderBeanFactory = modelElementMapperBuilderBeanFactory;
    }
}
